/*
 * WILDCMP.C
 *
 *	This is a simple '*' and '?' wildcard comparator.  We some relatively
 *	simple optimizations but the main one is to not repeatedly check a
 *	wildcard against portions of the string that have already been
 *	tested against the same wildcard and found to have failed.
 *
 * $Backplane: dspam/wildcmp.c,v 1.1.1.1 2003/04/22 09:21:14 dillon Exp $
 *
 * (c)Copyright 1997-2002, Matthew Dillon, All Rights Reserved.  Refer to
 *    the COPYRIGHT file in the base directory of this distribution 
 *    for specific rights granted.
 *
 */

#include "defs.h"

static int wildCmp(const char **mary, int d, const char *w, const char *s);
static int wildCaseCmp(const char **mary, int d, const char *w, const char *s);

int
WildCmp(const char *w, const char *s)
{
    int i;
    int c;
    int slen = strlen(s);
    const char **mary;

    for (i = c = 0; w[i]; ++i) {
	if (w[i] == '*')
	    ++c;
    }
    mary = calloc(sizeof(char *), c + 1);
    for (i = 0; i < c; ++i)
	mary[i] = s + slen;
    i = wildCmp(mary, 0, w, s);
    free(mary);
    return(i);
}

int
WildCaseCmp(const char *w, const char *s)
{
    int i;
    int c;
    int slen = strlen(s);
    const char **mary;

    for (i = c = 0; w[i]; ++i) {
	if (w[i] == '*')
	    ++c;
    }
    mary = calloc(sizeof(char *), c + 1);
    for (i = 0; i < c; ++i)
	mary[i] = s + slen;
    i = wildCaseCmp(mary, 0, w, s);
    free(mary);
    return(i);
}

/*
 * WildCmp() - compare wild string to sane string
 *
 *	Returns 0 on success, -1 on failure.
 */
static int
wildCmp(const char **mary, int d, const char *w, const char *s)
{
    int i;

    /*
     * skip fixed portion
     */
    for (;;) {
	switch(*w) {
	case '*':
	    /*
	     * optimize terminator
	     */
	    if (w[1] == 0)
		return(0);
	    if (w[1] != '?' && w[1] != '*') {
		/*
		 * optimize * followed by non-wild
		 */
		for (i = 0; s + i < mary[d]; ++i) {
		    if (s[i] == w[1] && wildCmp(mary, d + 1, w + 1, s + i) == 0)
			return(0);
		}
	    } else {
		/*
		 * less-optimal
		 */
		for (i = 0; s + i < mary[d]; ++i) {
		    if (wildCmp(mary, d + 1, w + 1, s + i) == 0)
			return(0);
		}
	    }
	    mary[d] = s;
	    return(-1);
	case '?':
	    if (*s == 0)
		return(-1);
	    ++w;
	    ++s;
	    break;
	default:
	    if (*w != *s)
		return(-1);
	    if (*w == 0)	/* terminator */
		return(0);
	    ++w;
	    ++s;
	    break;
	}
    }
    /* not reached */
    return(-1);
}


/*
 * WildCaseCmp() - compare wild string to sane string, case insensitive
 *
 *	Returns 0 on success, -1 on failure.
 */
static int
wildCaseCmp(const char **mary, int d, const char *w, const char *s)
{
    int i;

    /*
     * skip fixed portion
     */
    for (;;) {
	switch(*w) {
	case '*':
	    /*
	     * optimize terminator
	     */
	    if (w[1] == 0)
		return(0);
	    if (w[1] != '?' && w[1] != '*') {
		/*
		 * optimize * followed by non-wild
		 */
		for (i = 0; s + i < mary[d]; ++i) {
		    if (s[i] == w[1] && wildCaseCmp(mary, d + 1, w + 1, s + i) == 0)
			return(0);
		}
	    } else {
		/*
		 * less-optimal
		 */
		for (i = 0; s + i < mary[d]; ++i) {
		    if (wildCaseCmp(mary, d + 1, w + 1, s + i) == 0)
			return(0);
		}
	    }
	    mary[d] = s;
	    return(-1);
	case '?':
	    if (*s == 0)
		return(-1);
	    ++w;
	    ++s;
	    break;
	default:
	    if (*w != *s) {
		if (tolower(*w) != tolower(*s))
		    return(-1);
	    }
	    if (*w == 0)	/* terminator */
		return(0);
	    ++w;
	    ++s;
	    break;
	}
    }
    /* not reached */
    return(-1);
}

